#!/usr/bin/env -S python -u

import argparse
import logging
import pathlib

from sats_receiver import HOMEDIR, LOGSDIR, TLEDIR, RECDIR
from sats_receiver.manager import ReceiverManager
from sats_receiver.async_signal import AsyncSignal


def setup_logging(log_lvl):
    if not isinstance(log_lvl, int):
        raise ValueError('Invalid log level: %s' % log_lvl)

    fmt = '%(asctime)s %(levelname)s: %(name)s: %(message)s'
    logging.basicConfig(level=log_lvl, format=fmt, filename=LOGSDIR / 'sats_receiver.log')
    sh = logging.StreamHandler()
    sh.setFormatter(logging.Formatter(fmt))
    logging.getLogger().addHandler(sh)


if __name__ == '__main__':
    ap = argparse.ArgumentParser()
    ap.add_argument('config', type=pathlib.Path, help='Config file path')
    ap.add_argument('--log', default='INFO', type=(lambda x: getattr(logging, x.upper(), None)), help='Logging level, INFO default')
    ap.add_argument('--sysu', default=3600, type=int, help='System Usages info timeout in seconds, 1 hour default')
    args = ap.parse_args()

    for d in LOGSDIR, TLEDIR, RECDIR:
        if not d.exists():
            d.mkdir(parents=True, exist_ok=True)

    setup_logging(args.log)

    logging.info('Hello!')

    asig = AsyncSignal(['SIGABRT', 'SIGHUP', 'SIGINT', 'SIGTERM', 'SIGUSR1', 'SIGUSR2'])
    mng = ReceiverManager(args.config, args.sysu)

    while not mng.action():
        signame = asig.wait(1)
        if signame:
            if 'USR' in signame:
                # TODO
                pass
            else:
                mng.stop()
                logging.info('Exit by %s', signame)
                break

    mng.wait()
    logging.info('Bye!')



import atexit
import datetime as dt
import json
import logging
import multiprocessing as mp
import pathlib
import time
import queue

from typing import Mapping

from sats_receiver.gr_modules.receiver import RecUpdState, SatsReceiver
from sats_receiver.observer import Observer
from sats_receiver.tle import Tle
from sats_receiver.utils import Scheduler, SysUsage


class Executor(mp.Process):
    def __init__(self, sysu_intv=SysUsage.DEFAULT_INTV):
        self.prefix = self.__class__.__name__
        self.log = logging.getLogger(self.prefix)

        super().__init__(daemon=False)

        self.sysu_intv = sysu_intv
        self.q = mp.Queue()

    def run(self):
        self.log.debug('start')

        sysu = SysUsage(self.prefix, self.sysu_intv)

        while 1:
            sysu.collect()

            try:
                x = self.q.get(timeout=1)
            except queue.Empty:
                continue

            if x == '.':
                break

            try:
                fn, args, kwargs = x
            except ValueError:
                self.log.error('invalid task: %s', x)
                continue

            if callable(fn):
                try:
                    x = fn(*args, **kwargs)
                except Exception:
                    self.log.exception('%s with args=%s kwargs=%s', fn, args, kwargs)

                if x and isinstance(x, tuple) and len(x) == 4:
                    sat_name, fin_key, res_filename, end_time = x

        self.log.debug('finish')

    def execute(self, fn, *args, **kwargs):
        self.q.put((fn, args, kwargs))

    def stop(self):
        self.q.put('.')


class ReceiverManager:
    def __init__(self, config_filename: pathlib.Path, sysu_intv=SysUsage.DEFAULT_INTV, executor_cls=Executor):
        self.prefix = self.__class__.__name__
        self.log = logging.getLogger(self.prefix)

        self.sysu = SysUsage(self.prefix, sysu_intv)
        self.config_filename = config_filename
        self.config_file_stat = None
        self.config = {}

        self.receivers: dict[str, SatsReceiver] = {}
        self.stopped = False
        self.now = dt.datetime.now(dt.timezone.utc)
        self.file_failed_t = 0

        if not self.update_config(True, True):
            raise ValueError(f'{self.prefix}: Invalid config!')

        self.observer = Observer(self.config['observer'])
        self.tle = Tle(self.config['tle'])
        self.scheduler = Scheduler()
        self.executor = executor_cls(sysu_intv)
        self.executor.start()
        atexit.register(lambda x: (x.stop(), x.join()), self.executor)

        for cfg in self.config['receivers']:
            self._add_receiver(cfg)

    def _add_receiver(self, cfg: Mapping):
        if cfg.get('enabled', True):
            try:
                rec = SatsReceiver(self, cfg)
                self.receivers[cfg['name']] = rec
            except RuntimeError as e:
                self.log.error('Skip receiver "%s": %s', cfg['name'], e)
        else:
            self.log.debug('Skip disabled receiver `%s`', cfg['name'])

    @property
    def t(self) -> dt.datetime:
        """
        Set and return current time
        """

        self.now = dt.datetime.now(dt.timezone.utc)
        return self.now

    def stop(self):
        for rec in self.receivers.values():
            rec.stop()

        self.executor.stop()
        self.stopped = True

        self.log.info('finish')

    def wait(self):
        for rec in self.receivers.values():
            rec.wait()
        self.executor.join()

    def update_config(self, init=False, force=False):
        """
        :param init: True when called from __init__
        :param force: True if you need to force update
        :return: True if config update success
        """

        needs = any(i.updated != RecUpdState.NO_NEED
                    for i in self.receivers.values())

        if not (self._check_config() or force or needs):
            return

        try:
            new_cfg = json.load(self.config_filename.open())
        except (IOError, json.JSONDecodeError) as e:
            self.log.error('Error during load config: %s', e)
            return

        if not (force or needs) and new_cfg == self.config:
            return

        if not self._validate_config(new_cfg):
            self.log.warning('invalid new config!')
            return

        if new_cfg != self.config:
            self.log.debug('reconf')
            self.config = new_cfg

        if init:
            return 1

        self.observer.update_config(self.config['observer'])
        self.tle.update_config(self.config['tle'])

        for cfg in self.config['receivers']:
            receiver = self.receivers.get(cfg['name'])
            if receiver:
                if ((force or receiver.updated == RecUpdState.FORCE_NEED)
                        or (not receiver.is_runned and receiver.updated != RecUpdState.NO_NEED)):
                    if not cfg.get('enabled', True):
                        self.log.debug('%s: stop by disabling', receiver.name)
                        receiver.stop()
                        receiver.wait()
                        self.receivers.pop(receiver.name)
                        continue

                    try:
                        receiver.update_config(cfg, force)
                    except RuntimeError as e:
                        receiver.stop()
                        receiver.wait()
                        self.receivers.pop(receiver.name)
                        self.log.error('%s: cannot update config: %s. Stop', receiver.name, e)

            else:
                self._add_receiver(cfg)

        return 1

    def _check_config(self):
        try:
            st = self.config_filename.stat()
        except FileNotFoundError:
            t = time.monotonic()
            if t - self.file_failed_t > 300:
                self.log.error('Config file does\'t exist: %s', self.config_filename)
                self.file_failed_t = t
        else:
            old = self.config_file_stat
            if not old or (st.st_ino != old.st_ino
                           or st.st_dev != old.st_dev
                           or st.st_mtime != old.st_mtime
                           or st.st_size != old.st_size
                           or st.st_mode != old.st_mode):
                self.config_file_stat = st
                for receiver in self.receivers.values():
                    receiver.updated = RecUpdState.UPD_NEED

                return 1

    @staticmethod
    def _validate_config(config: Mapping) -> bool:
        return all(map(lambda x: x in config, [
            'observer',
            'tle',
            'receivers',
        ]))

    def action(self):
        self.sysu.collect()

        if self.stopped:
            return 1

        try:
            self.update_config()
            self.scheduler.action()
            recalc = self.observer.action(self.t)
            need_upd = self.tle.action(self.now)

            for receiver in self.receivers.values():
                if recalc:
                    receiver.recalculate_pass()

                if need_upd:
                    receiver.updated = RecUpdState.UPD_NEED

                receiver.action()

        except Exception as e:
            self.log.exception('%s. Exit', e)
            self.stop()



import logging
import math

from typing import Union

import gnuradio as gr
import gnuradio.analog
import gnuradio.digital
import gnuradio.filter
import gnuradio.gr


class QpskDemod(gr.gr.hier_block2):
    def __init__(self,
                 samp_rate: Union[int, float],
                 baudrate: Union[int, float],
                 excess_bw: Union[int, float] = None,
                 ntaps: int = None,
                 costas_bw: Union[int, float] = None):
        super(QpskDemod, self).__init__(
            'QPSK Demodulator',
            gr.gr.io_signature(1, 1, gr.gr.sizeof_gr_complex),
            gr.gr.io_signature(1, 1, gr.gr.sizeof_float)
        )

        if excess_bw is None:
            excess_bw = 0.35
        if ntaps is None:
            ntaps = 33
        if costas_bw is None:
            costas_bw = 0.005

        self.samp_rate = samp_rate
        self.baudrate = baudrate
        self.excess_bw = excess_bw
        self.ntaps = int(ntaps)

        self.rrc = gr.filter.fir_filter_ccf(
            1,
            gr.filter.firdes.root_raised_cosine(
                gain=1,
                sampling_freq=samp_rate,
                symbol_rate=baudrate,
                alpha=excess_bw,
                ntaps=ntaps
            )
        )
        self.agc = gr.analog.agc_cc(0.1, 1.0, 1.0)
        self.agc.set_max_gain(10e6)
        self.costas = gr.digital.costas_loop_cc(costas_bw, 4, False)
        # self.symbol_sync = gr.digital.symbol_sync_cc(
        #     detector_type=gr.digital.TED_MUELLER_AND_MULLER,
        #     sps=samp_rate / baudrate,
        #     loop_bw=(2 * math.pi) / (2 * baudrate),
        #     damping_factor=1.0,
        #     ted_gain=1.0,
        #     max_deviation=1.5,
        #     osps=1,
        #     slicer=gr.digital.constellation_qpsk().base(),
        #     interp_type=gr.digital.IR_MMSE_8TAP,
        #     n_filters=128,
        #     taps=[],
        # )
        self.recov = gr.digital.clock_recovery_mm_cc(
            omega=samp_rate / baudrate,
            gain_omega=1e-6,
            mu=0.5,
            gain_mu=0.01,
            omega_relative_limit=0.01,
        )
        self.constel_decoder = gr.digital.constellation_soft_decoder_cf(gr.digital.constellation_qpsk().base())

        self.connect(
            self,
            self.rrc,
            self.agc,
            self.costas,
            # self.symbol_sync,
            self.recov,
            self.constel_decoder,
            self,
        )



import datetime as dt
import logging
import math
import pathlib

from typing import Optional, Union

import dateutil.tz
import ephem
import gnuradio as gr
import gnuradio.analog
import gnuradio.blocks
import gnuradio.digital
import gnuradio.fft
import gnuradio.filter
import gnuradio.gr

from sats_receiver import utils
from sats_receiver.systems import apt


class Decoder(gr.gr.hier_block2):
    def __init__(self,
                 name: str,
                 sat_name: str,
                 samp_rate: Union[int, float],
                 out_dir: pathlib.Path):
        self.prefix = f'{self.__class__.__name__}: {sat_name}'
        self.log = logging.getLogger(self.prefix)

        super(Decoder, self).__init__(
            name,
            gr.gr.io_signature(1, 1, gr.gr.sizeof_gr_complex),
            gr.gr.io_signature(0, 0, 0)
        )

        self.now = dt.datetime.fromtimestamp(0)
        self.sat_name = sat_name
        self.samp_rate = samp_rate
        self.out_dir = out_dir
        self.tmp_file = out_dir / ('_'.join(name.lower().split()) + '.tmp')
        self.base_kw = dict(log=self.log, sat_name=sat_name, out_dir=out_dir)

    def start(self):
        self.tmp_file = self.out_dir / ('_'.join([self.t.strftime('%Y%m%d%H%M%S'),
                                                  *self.name().lower().split()]) + '.tmp')
        self.base_kw.update(tmp_file=self.tmp_file)

    def finalize(self, executor, fin_key: str):
        pass

    @property
    def t(self) -> dt.datetime:
        t = dt.datetime.now()
        self.now = t
        return t


class RawDecoder(Decoder):
    def __init__(self,
                 sat_name: str,
                 samp_rate: Union[int, float],
                 out_dir: pathlib.Path):
        super(RawDecoder, self).__init__('Raw Decoder', sat_name, samp_rate, out_dir)

        self.ctf = gr.blocks.complex_to_float(1)
        self.wav_sink = gr.blocks.wavfile_sink(
            str(self.tmp_file),
            2,
            samp_rate,
            gr.blocks.FORMAT_WAV,
            gr.blocks.FORMAT_FLOAT,
            False
        )
        self.wav_sink.close()
        self.tmp_file.unlink(True)

        self.connect(self, self.ctf, self.wav_sink)
        self.connect((self.ctf, 1), (self.wav_sink, 1))

    def start(self):
        super(RawDecoder, self).start()

        self.wav_sink.open(str(self.tmp_file))

    def finalize(self, executor, fin_key: str):
        self.wav_sink.close()
        if self.tmp_file.exists():
            executor.execute(self._raw_finalize, **self.base_kw, fin_key=fin_key)

    @staticmethod
    def _raw_finalize(log, sat_name, out_dir, tmp_file, fin_key) -> Optional[tuple[str, str, str, dt.datetime]]:
        log.debug('finalizing...')

        d = dt.datetime.fromtimestamp(tmp_file.stat().st_mtime, dateutil.tz.tzutc())
        res_fn = tmp_file.rename(out_dir / d.strftime(f'{sat_name}_%Y-%m-%d_%H-%M-%S_RAW.wav'))
        st = res_fn.stat()
        log.info('finish: %s (%s)', res_fn, utils.numdisp(st.st_size))

        return sat_name, fin_key, res_fn, dt.datetime.fromtimestamp(st.st_mtime, dateutil.tz.tzutc())


class AptDecoder(Decoder):
    def __init__(self,
                 sat_name: str,
                 samp_rate: Union[int, float],
                 out_dir: pathlib.Path,
                 sat_ephem_tle: Optional[tuple[ephem.EarthSatellite, tuple[str, str, str]]],
                 observer_lonlat: tuple[float, float]):
        name = 'APT Decoder'
        super(AptDecoder, self).__init__(name, sat_name, samp_rate, out_dir)

        self.corr_file = out_dir / ('_'.join(name.lower().split()) + '.corr')
        self.peaks_file = out_dir / ('_'.join(name.lower().split()) + '.peaks')
        self.sat_ephem_tle = sat_ephem_tle
        self.observer_lonlat = observer_lonlat
        self.base_kw.update(sat_tle=self.sat_ephem_tle[1], observer_lonlat=self.observer_lonlat)

        resamp_gcd = math.gcd(samp_rate, apt.Apt.WORK_RATE)

        self.frs = gr.blocks.rotator_cc(2 * math.pi * -apt.Apt.CARRIER_FREQ / samp_rate)
        self.rsp = gr.filter.rational_resampler_ccc(
            interpolation=apt.Apt.WORK_RATE // resamp_gcd,
            decimation=samp_rate // resamp_gcd,
            taps=[],
            fractional_bw=0,
        )

        tofilter_freq = 2080
        lp_gcd = math.gcd(apt.Apt.WORK_RATE, tofilter_freq)

        self.lpf = gr.filter.fir_filter_ccf(
            1,
            gr.filter.firdes.low_pass(
                gain=1,
                sampling_freq=apt.Apt.WORK_RATE / lp_gcd,
                cutoff_freq=tofilter_freq / lp_gcd,
                transition_width=0.1,
                window=gr.fft.window.WIN_HAMMING,
                param=6.76,
            )
        )
        self.ctm = gr.blocks.complex_to_mag()
        self.ftc = gr.blocks.float_to_complex()
        # self.dc_rem = gr.blocks.correctiq()
        self.correllator = gr.digital.corr_est_cc(
            symbols=apt.Apt.SYNC_A,
            sps=1,
            mark_delay=1,
            threshold=0.9,
            threshold_method=gr.digital.THRESHOLD_ABSOLUTE,
        )
        self.ctf_out = gr.blocks.complex_to_float()
        self.ctf_corr = gr.blocks.complex_to_float()

        self.out_file_sink = gr.blocks.file_sink(gr.gr.sizeof_float, str(self.tmp_file), False)
        self.out_file_sink.close()
        self.tmp_file.unlink(True)

        self.out_corr_sink = gr.blocks.file_sink(gr.gr.sizeof_float, str(self.corr_file), False)
        self.out_corr_sink.close()
        self.corr_file.unlink(True)

        self.peak_detector = gr.blocks.peak_detector2_fb(
            threshold_factor_rise=7,
            look_ahead=apt.Apt.FRAME_WIDTH * apt.Apt.WORK_RATE // apt.Apt.FINAL_RATE,
            alpha=0.001,
        )
        self.out_peaks_sink = gr.blocks.file_sink(gr.gr.sizeof_char, str(self.peaks_file), False)
        self.out_peaks_sink.close()
        self.peaks_file.unlink(True)

        self.connect(
            self,
            self.frs,
            self.rsp,
            self.lpf,
            self.ctm,
            self.ftc,
            # self.dc_rem,
            self.correllator,
            self.ctf_out,
            self.out_file_sink,
        )
        self.connect(
            (self.correllator, 1),
            self.ctf_corr,
            self.out_corr_sink,
        )
        self.connect(
            self.ctf_corr,
            self.peak_detector,
            self.out_peaks_sink,
        )

    def start(self):
        super(AptDecoder, self).start()
        self.corr_file = self.out_dir / ('_'.join([self.now.strftime('%Y%m%d%H%M%S'),
                                                   *self.name().lower().split()]) + '.corr.tmp')
        self.peaks_file = self.out_dir / ('_'.join([self.now.strftime('%Y%m%d%H%M%S'),
                                                    *self.name().lower().split()]) + '.peaks.tmp')
        self.base_kw.update(corr_file=self.corr_file, peaks_file=self.peaks_file)

        self.out_file_sink.open(str(self.tmp_file))
        self.out_file_sink.set_unbuffered(False)

        self.out_corr_sink.open(str(self.corr_file))
        self.out_corr_sink.set_unbuffered(False)

        self.out_peaks_sink.open(str(self.peaks_file))
        self.out_peaks_sink.set_unbuffered(False)

    def finalize(self, executor, fin_key: str):
        self.out_file_sink.close()
        self.out_corr_sink.close()
        self.out_peaks_sink.close()

        for p in self.tmp_file, self.corr_file, self.peaks_file:
            if not p.exists():
                self.log.warning('%s: missing components `%s`', p)
                self.tmp_file.unlink(True)
                self.corr_file.unlink(True)
                self.peaks_file.unlink(True)
                return

        executor.execute(self._apt_finalize, **self.base_kw, fin_key=fin_key)

    @staticmethod
    def _apt_finalize(log: logging.Logger,
                      sat_name: str,
                      sat_tle: tuple[str, str, str],
                      observer_lonlat: tuple[float, float],
                      tmp_file: pathlib.Path,
                      corr_file: pathlib.Path,
                      peaks_file: pathlib.Path,
                      out_dir: pathlib.Path,
                      fin_key: str) -> Optional[tuple[str, str, pathlib.Path, dt.datetime]]:
        log.debug('finalizing...')

        a = apt.Apt(sat_name, tmp_file, corr_file, peaks_file, sat_tle, observer_lonlat)
        x = a.process()

        tmp_file.unlink(True)
        corr_file.unlink(True)
        peaks_file.unlink(True)

        if x:
            log.info('finish with error')
        else:
            res_fn, sz = a.to_apt(out_dir)
            log.info('finish: %s (%s)', res_fn, utils.numdisp(sz))

            return sat_name, fin_key, res_fn, a.end_time


class RawStreamDecoder(Decoder):
    def __init__(self,
                 sat_name: str,
                 samp_rate: Union[int, float],
                 out_dir: pathlib.Path,
                 name='RAW Stream Decoder'):
        super(RawStreamDecoder, self).__init__(name, sat_name, samp_rate, out_dir)

        self.ctf = gr.blocks.complex_to_float(1)
        self.rail = gr.analog.rail_ff(-1, 1)
        self.ftch = gr.blocks.float_to_char(1, 127)
        # self.fts = gr.blocks.float_to_short(1, 32767)
        # self.stch = gr.blocks.short_to_char(1)

        self.out_file_sink = gr.blocks.file_sink(gr.gr.sizeof_char, str(self.tmp_file), False)
        self.out_file_sink.close()
        self.tmp_file.unlink(True)

        self.connect(
            self,
            self.ctf,
            self.rail,
            self.ftch,
            self.out_file_sink,
        )

    def start(self):
        super(RawStreamDecoder, self).start()

        self.out_file_sink.open(str(self.tmp_file))
        self.out_file_sink.set_unbuffered(False)

    def finalize(self, executor, fin_key: str):
        self.out_file_sink.close()
        if self.tmp_file.exists():
            executor.execute(self._raw_stream_finalize, **self.base_kw, fin_key=fin_key)

    @staticmethod
    def _raw_stream_finalize(log: logging.Logger,
                             sat_name: str,
                             out_dir: pathlib.Path,
                             tmp_file: pathlib.Path,
                             fin_key: str) -> Optional[tuple[str, str, pathlib.Path, dt.datetime]]:
        log.debug('finalizing...')

        d = dt.datetime.fromtimestamp(tmp_file.stat().st_mtime, dateutil.tz.tzutc())
        res_fn = tmp_file.rename(out_dir / d.strftime(f'{sat_name}_%Y-%m-%d_%H-%M-%S_RAW.s'))
        st = res_fn.stat()
        log.info('finish: %s (%s)', res_fn, utils.numdisp(st.st_size))

        return sat_name, fin_key, res_fn, dt.datetime.fromtimestamp(st.st_mtime, dateutil.tz.tzutc())


class LrptDecoder(RawStreamDecoder):
    def __init__(self,
                 sat_name: str,
                 samp_rate: Union[int, float],
                 out_dir: pathlib.Path):
        raise NotImplementedError()
        super(LrptDecoder, self).__init__(sat_name, samp_rate, out_dir, name='LRPT Decoder')

    def start(self):
        super(LrptDecoder, self).start()

    def finalize(self, executor, fin_key: str):
        super(LrptDecoder, self).finalize(executor, fin_key)



import datetime as dt
import dateutil.tz
import enum
import logging
import pathlib

from typing import Mapping, Union

import gnuradio as gr
import gnuradio.blocks
import gnuradio.gr
import gnuradio.soapy

from sats_receiver.gr_modules import modules

try:
    from sats_receiver import librtlsdr
    LIBRTLSDR = None
except OSError as e:
    LIBRTLSDR = str(e)


class RecUpdState(enum.IntEnum):
    UPD_NEED = enum.auto()
    FORCE_NEED = enum.auto()
    NO_NEED = enum.auto()


class SatsReceiver(gr.gr.top_block):
    def __init__(self, up, config: Mapping):
        n = config.get('name', '')
        self.prefix = f'Receiver{n and f": {n}"}'
        self.log = logging.getLogger(self.prefix)

        super(SatsReceiver, self).__init__('Sats Receiver', catch_exceptions=True)

        self.up = up
        self.config = {}
        self.satellites: dict[str, modules.Satellite] = {}
        self.is_runned = False
        self.updated = RecUpdState.UPD_NEED

        self.signal_src = gr.blocks.null_source(gr.gr.sizeof_gr_complex)
        self.blocks_correctiq = gr.blocks.correctiq()
        self.src_null_sink = gr.blocks.null_sink(gr.gr.sizeof_gr_complex)

        if not self.update_config(config):
            raise ValueError(f'{self.prefix}: Invalid config!')

    def update_config(self, config: Mapping, force=False):
        """
        :param config: new config
        :param force: True if you need to force update
        :return: True if config update success
        """

        if force or self.updated != RecUpdState.NO_NEED or self.config != config:
            if not self._validate_config(config):
                self.log.warning('invalid new config!')
                return

            if self.is_runned and not config.get('enabled', True):
                self.log.debug('stop by disabling')
                self.stop()
                self.wait()

            if (self.is_runned
                    and (self.source != config['source']
                         or self.serial != config.get('serial', ''))):
                self.log.warning('Could not change receiver on running state!')
                return

            try:
                old_src, old_serial = self.source, self.serial
            except KeyError:
                old_src, old_serial = '', ''

            self.config = config

            if old_src != self.source or old_serial != self.serial:
                gr.soapy.source(f'driver={self.source}{self.serial and f",serial={self.serial}"}',
                                'fc32', 1, '', '', [''], [''])
                self.log.info('New source found')

            if self.is_runned:
                self.signal_src.set_sample_rate(0, self.samp_rate)
                self.signal_src.set_frequency(0, self.tune)
                self.signal_src.set_frequency_correction(0, 0)
                self.signal_src.set_gain_mode(0, False)
                self.signal_src.set_gain(0, 'TUNER', self.gain)
            else:
                self.set_biast(0)

            exist_sats = set(self.satellites.keys())
            new_cfg_sats = {i['name']: i for i in self.sats}
            new_cfg_sats_names = set(new_cfg_sats.keys())

            to_remove_sats = exist_sats - new_cfg_sats_names
            to_update_sats = exist_sats & new_cfg_sats_names
            to_create_sats = new_cfg_sats_names - exist_sats

            for sat_name in to_update_sats:
                sat = self.satellites[sat_name]
                if sat.config != new_cfg_sats:
                    to_remove_sats.add(sat_name)
                    to_create_sats.add(sat_name)

            if to_remove_sats or to_create_sats:
                self.lock()

            for sat_name in to_remove_sats:
                sat = self.satellites[sat_name]
                self.up.scheduler.cancel(*sat.events)
                sat.stop()
                if self.is_runned:
                    self.disconnect(self.blocks_correctiq, sat)
                del self.satellites[sat_name]

            for sat_name in to_create_sats:
                cfg = new_cfg_sats[sat_name]
                if not cfg.get('enabled', True):
                    self.log.debug('Skip disabled sat `%s`', sat_name)
                    continue

                sat_ephem_tle = self.up.tle.get(sat_name)
                if sat_ephem_tle is None:
                    self.log.info('Sat `%s` not found in TLE. Skip', sat_name)
                    continue

                try:
                    sat = modules.Satellite(cfg, sat_ephem_tle, self.up.observer.lonlat, self.tune, self.samp_rate, self.output_directory, self.up.executor)
                except ValueError as e:
                    self.log.warning('%s: %s. Skip', sat_name, e)
                    continue

                if self.calculate_pass(sat):
                    if self.is_runned:
                        self.connect(self.blocks_correctiq, sat)
                    self.satellites[sat.name] = sat

            if to_remove_sats or to_create_sats:
                self.unlock()

            self.updated = RecUpdState.NO_NEED

            return 1

    @staticmethod
    def _validate_config(config: Mapping) -> bool:
        return all(map(lambda x: x in config, [
            'name',
            # 'enabled',  # optional
            'source',
            # 'serial',   # optional
            # 'biast',    # optional
            # 'gain',     # optional
            'tune',
            'samp_rate',
            'output_directory',
            'sats',
        ]))

    @property
    def name(self) -> str:
        return self.config['name']

    @property
    def enabled(self) -> bool:
        return self.config.get('enabled', True)

    @property
    def source(self) -> str:
        return self.config['source']

    @property
    def serial(self) -> str:
        return self.config.get('serial', '')

    @property
    def biast(self) -> bool:
        return self.config.get('biast', False)

    @property
    def gain(self) -> Union[int, float]:
        """
        Receiver gain, db
        """

        return self.config.get('gain', 0)

    @property
    def tune(self) -> Union[int, float]:
        """
        Receiver tune frequency, Hz
        """

        return self.config['tune']

    @property
    def samp_rate(self) -> Union[int, float]:
        """
        Receiver samplerate, Hz
        """

        return self.config['samp_rate']

    @property
    def output_directory(self) -> pathlib.Path:
        return pathlib.Path(self.config['output_directory']).expanduser()

    @property
    def sats(self) -> list[Mapping]:
        return self.config['sats']

    @property
    def is_active(self) -> bool:
        """
        True when at least one satellite recorder is running
        """

        return any(x.is_runned for x in self.satellites.values())

    def start(self, max_noutput_items=10000000):
        if self.enabled and not self.is_runned:
            self.log.info('START tune=%s samp_rate=%s gain=%s biast=%s',
                          self.tune, self.samp_rate, self.gain, self.biast)

            self.set_biast(self.biast)

            try:
                self.signal_src = gr.soapy.source(f'driver={self.source}{self.serial and f",serial={self.serial}"}',
                                                  'fc32', 1, '', '', [''], [''])
            except RuntimeError as e:
                self.log.error('cannot start: %s', e)

                self.stop()
                self.wait()
                self.set_biast(0, True)

                t = self.up.now + dt.timedelta(minutes=5)
                for sat in self.satellites.values():
                    self.up.scheduler.cancel(*sat.events)
                    if sat.enabled:
                        self.up.scheduler.plan(t, self.calculate_pass, sat)

                return 1

            self.signal_src.set_sample_rate(0, self.samp_rate)
            self.signal_src.set_frequency(0, self.tune)
            self.signal_src.set_frequency_correction(0, 0)
            self.signal_src.set_gain_mode(0, False)
            self.signal_src.set_gain(0, 'TUNER', self.gain)

            self.connect(
                self.signal_src,
                self.blocks_correctiq,
                self.src_null_sink,
            )

            for sat in self.satellites.values():
                self.connect(self.blocks_correctiq, sat)

            super(SatsReceiver, self).start(max_noutput_items)
            self.is_runned = True

    def stop(self, sched_clear=True):
        """
        Stop the flowgraph

        :param sched_clear: clear scheduled tasks
        """
        if self.is_runned:
            self.log.info('STOP')

            super(SatsReceiver, self).stop()

            self.disconnect(
                self.signal_src,
                self.blocks_correctiq,
                self.src_null_sink,
            )
            self.signal_src = gr.blocks.null_source(gr.gr.sizeof_gr_complex)

        for sat in self.satellites.values():
            if sched_clear:
                self.up.scheduler.cancel(*sat.events)
            sat.stop()
            if self.is_runned:
                self.disconnect(self.blocks_correctiq, sat)

        self.is_runned = False

    def action(self):
        if self.is_active and not self.start():
            for sat in self.satellites.values():
                sat.correct_doppler(self.up.observer.get_obj())
        elif self.is_runned:
            self.stop(False)
            self.wait()
            self.set_biast(0)

    def calculate_pass(self, sat: modules.Satellite):
        """
        Calculate satellite pass

        :return: True when calculate success
        """

        x = self.up.tle.get_ephem(sat.name)
        if x:
            t = self.up.now
            tt = t + dt.timedelta(hours=24)
            ltz = dateutil.tz.tzlocal()

            while t <= tt:
                try:
                    rise_t, rise_az, culm_t, culm_alt, set_t, set_az = self.up.observer.next_pass(x, t)
                except ValueError as e:
                    # skip circumpolar/non-setting satellite
                    self.log.error('Sat `%s` skip: %s', e)
                    return

                set_tt = set_t + dt.timedelta(seconds=5)
                if culm_alt >= sat.min_elevation:
                    if set_t < rise_t:
                        rise_t = t
                    sat.events = [
                        None if sat.is_runned else self.up.scheduler.plan(rise_t, sat.start),
                        self.up.scheduler.plan(set_t, sat.stop),
                        self.up.scheduler.plan(set_tt, self.calculate_pass, sat)
                    ]
                    self.log.info('Sat `%s` planned on %s <-> %s', sat.name, rise_t.astimezone(ltz), set_t.astimezone(ltz))
                    break

                t = set_tt

            if t > tt:
                self.log.info('Sat `%s`: No passes found for the next 24 hours', sat.name)
                self.up.scheduler.plan(tt, self.calculate_pass, sat)

            return 1

        self.log.info('Sat `%s` not found in TLE. Skip', sat.name)

    def recalculate_pass(self):
        for sat in self.satellites.values():
            self.up.scheduler.cancel(*sat.events)
            sat.events = [None, None, None]
            self.calculate_pass(sat)

    def set_biast(self, v, silent=False):
        if not LIBRTLSDR and self.source == 'rtlsdr':
            try:
                librtlsdr.set_bt(v, self.serial)
            except librtlsdr.LibRtlSdrError as e:
                if not silent:
                    self.log.info('change bias-t error: %s', e)



import pathlib
import sys

from setuptools import setup, find_packages


# Check python version
MINIMAL_PY_VERSION = (3, 7)
if sys.version_info < MINIMAL_PY_VERSION:
    raise RuntimeError('This app works only with Python {}+'.format('.'.join(map(str, MINIMAL_PY_VERSION))))

HERE = pathlib.Path(__file__).parent


def get_version():
    for line in (HERE / 'sats_receiver' / '__init__.py').open():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]


def get_long_description():
    return (HERE / 'README.md').read_text('utf-8')


setup(
    name='sats_receiver',
    version=get_version(),
    url='https://github.com/baskiton/sats-receiver',
    project_urls={
        'Bug Tracker': "https://github.com/baskiton/sats-receiver/issues",
    },
    license='MIT',
    author='Alexander Baskikh',
    python_requires='>=3.7',
    author_email='baskiton@gmail.com',
    description='Satellites data receiver based on GNU Radio',
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=[
        'ephem',
        'numpy',
        'Pillow',
        'pyshp',
        'python-dateutil',
        'scipy',
    ],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Environment :: Console',
        'Programming Language :: Python :: 3',
        'Topic :: Communications :: Ham Radio',
    ]
)



import errno
import fcntl
import os
import select
import signal
import time


SIG = {}
for n, v in signal.__dict__.items():
    if n.startswith('SIG') and n.isalnum() and isinstance(v, int):
        SIG[n] = v
        SIG[n[3:]] = v
        SIG[v] = n


class AsyncSignal:
    def __init__(self, signals=None):
        super(AsyncSignal, self).__init__()

        self.olds = {}

        if signals is None:
            signals = ['SIGINT']

        for s in signals:
            try:
                sig = int(s)
            except ValueError:
                sig = SIG[s]

            assert 0 < sig < 256 and sig in SIG

            old = signal.signal(sig, self._handler)
            if old is None:
                old = signal.SIG_DFL
            self.olds[sig] = old

        if self.olds:
            self.pid = os.getpid()
            self.rd, self.wr = os.pipe()
            self._tune(self.rd)
            self._tune(self.wr)

    def close(self):
        if self.olds:
            for sig, old in self.olds.items():
                signal.signal(sig, old)
            os.close(self.rd)
            os.close(self.wr)
        self.olds = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _tune(self, fd):
        fcntl.fcntl(fd, fcntl.F_SETFL, fcntl.fcntl(fd, fcntl.F_GETFL) | os.O_NONBLOCK)
        fcntl.fcntl(fd, fcntl.F_SETFD, fcntl.fcntl(fd, fcntl.F_GETFD) | fcntl.FD_CLOEXEC)

    def _handler(self, signum, frame):
        if os.getpid() != self.pid:
            os.kill(self.pid, signum)
        else:
            try:
                os.write(self.wr, chr(signum).encode())
            except ValueError:
                pass

    def wait(self, t=None):
        if t and t > 0:
            till = time.monotonic() + t

        while 1:
            try:
                select.select((self.rd,), (), (), t)
                break
            except select.error as e:
                if e[0] == errno.EINTR:
                    if t and t > 0:
                        t = till - time.monotonic()
                        if t <= 0:
                            break
                    continue
                if e[0] == errno.EBADF:
                    return
                raise

        try:
            return SIG[ord(os.read(self.rd, 1))]
        except:
            return
